function [nodes_or_indices, param_or_coeffs] = entryId_cs(N, m_1, m_2, s, d, basis_types, return_type, funcEval, tau, param_nodes1d)
%ENTRYID_CS performs entry identifications using compressed sensing approach.
%  Supports Fourier, (preconditioned) Legendre, and Chebyshev basis
%    (also mixed bases).
%
%  Input:
%    N - scalar, number of possible basis indices per spatial dimension
%        (side length of cube)
%    m_1 - scalar, number of samples for prefix dimensions (first part, w)
%    m_2 - scalar, number of samples for prefix dimensions (second part, z)
%    s - scalar, sparsity
%    d - scalar, spatial dimension
%    basis_types - cell array of length d,
%                  contains either 'fourier' or 'chebyshev' or 'legendre'
%    return_type - either 'nodes' or indices', determines output variables,
%                  'nodes' returns pairing grids, 'indices' performs pairing
%    funcEval - (complex) double matrix: (m_1 * m_2) \times 1,
%               contains function samples evaluated at required nodes,
%               only used for return_type='indices'
%    tau - scalar >= 0, threshold parameter for entry identification
%    param_nodes1d - cell array of length d containing
%                    double matrix: m_1 \times 1
%
%  Output:
%    nodes_or_indices - for return_type='nodes': cell array of length d
%                       containing double matrix: (m_1 * m_2) \times d,
%                       for return_type='indices': cell array of length d
%                       containing column vector of detected basis indices
%    param_or_coeffs - for return_type='nodes': param_nodes1d,
%                      for return_type='indices: cell array of length d
%                      containing column vector of estimated coefficients
%
% version with speed improvements and generalizations 2018-2019 by Toni Volkmer

switch return_type
  case 'nodes'
    [nodes_or_indices, param_or_coeffs] = entryId_cs_nodes(m_1, m_2, d, basis_types);
  case 'indices'
    [nodes_or_indices, param_or_coeffs] = entryId_cs_indices(N, m_1, m_2, s, d, basis_types, funcEval, tau, param_nodes1d);
  otherwise
    error('Unknown return_type: %s', return_type);
end

end

function [indices, coeffs] = entryId_cs_indices(N, m_1, m_2, s, d, basis_types, funcEval, tau, param_nodes1d)
  indices = cell(d,1);
  coeffs = cell(d,1);
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  for j=1:d
    funcEval_j = reshape(funcEval{j},m_1,m_2);
    indices{j}=[]; %entry set N_j in the paper
    coeffs{j}=[];
    possible_indices_j = (0:N-1)';
    basisEval = basis_eval(basis_types(j), possible_indices_j, param_nodes1d{j});
    basisEval = basisEval.';
    energy_est = sum((abs(conj(basisEval(:,:))*funcEval_j(:,:))).^2, 2);
    energy_est=energy_est/(m_1*(m_1-1)*m_2); %normalization
    % sorting to get the updated P_j
    [sortE,I]=sort(energy_est,'descend');%sorting according to the size of E_{j,(\tilde{n}, ...)}
    leng = length(possible_indices_j);
    if leng<s
        ss=leng; % for the case where leng is smaller than s 
    else
        ss=s; 
    end    
    found_indices_j = possible_indices_j(I(1:ss),:);
    found_indices_j = found_indices_j(sortE(1:ss) >= tau^2);
    indices{j} = found_indices_j;
  end
end

function [nodes, param_nodes1d] = entryId_cs_nodes(m_1, m_2, d, basis_types)
  nodes = cell(d,1);
  param_nodes1d = cell(d,1);
  x = basis_gen_rnd_nodes(basis_types, m_2);
  xprime = basis_gen_rnd_nodes(basis_types, m_1);
  for j = 1:d
    xxprime_j = nan(m_1,m_2,d);
    for k=1:m_1
      xxprime_j(k,:,:) = x;
    end
    param_nodes1d{j} = xprime(:,j);
    xxprime_j(:,:,j) = repmat(xprime(:,j),1,m_2);
    nodes{j} = reshape(xxprime_j,m_1*m_2,d);
  end
end
