function [a, supp, t, num_iter] = sublin_cosamp_v2(samples_ls, nodes_ls, samples_entryid, nodes_entryid, entryid_fct, entryid_fct_params, samples_pairing, nodes_w_pairing, nodes_wz_pairing, pairing_fct, N, s, cosamp_max_iter, basis_types)
%SUBLIN_COSAMP_V2 sublinearized CoSAMP, generalized and improved version 2.
%  This is an implementation of Algorithm 1 and 2.
%
%  Input:
%    samples_ls - samples for least squares coefficients computations
%    nodes_ls - corresponding nodes
%    samples_entryid - samples for entry identification
%    nodes_entryid - corresponding nodes
%    entryid_fct - function handle of entry identification function
%    entryid_fct_params - corresponding internal parameters
%    samples_pairing - samples for pairing step
%    nodes_w_pairing - prefix nodes w_{j,\ell}
%    nodes_wz_pairing - all pairing nodes
%    pairing_fct - function handle of pairing function
%    s - scalar, sparsity
%    cosamp_max_iter - scalar >= 1, termination criterion for algorithm
%    basis_types - cell array of length d,
%                  contains either 'fourier' or 'chebyshev' or 'legendre'
%                  or 'legendre_preconditioned'
%
%  Output:
%    a - coefficients belonging to detected basis indices
%    supp - detected basis indices
%    t - iterative returned
%    num_iter - number of iterations performed
%
% based on SHT implementation 2018 by Bosu Choi
% version with speed improvements and generalizations 2018-2019 by Toni Volkmer

d = size(nodes_ls,2);
m_bar = size(nodes_ls,1);

a = zeros(0,1); % initilization of s-sparse approximation
t = 0; %iterative
num_iter = 0; %iterations performed
supp = zeros(0,d); % initialization of the support set S
supp_old = supp;
residual_norm = norm(samples_ls); % l2 norm of the residual vector

v_i = samples_entryid; % initialization of the samples for entryId
v_p = samples_pairing; % initialization of the samples for pairing

show_warning = 0;

stagnation_counter = 0;
% main loop
while t < cosamp_max_iter
  if (t > 0) && (size(intersect(supp,supp_old,'rows'),1) == size(supp,1))
    stagnation_counter = stagnation_counter + 1;
    if stagnation_counter >= 2
      if show_warning
        warning('Aborting algorithm after two consecutive iterations of identical basis indices');
      end
      break;
    end
  else
    stagnation_counter = 0;
  end
  
  
  t=t+1; %updating the iteration number
  num_iter = t;
  %entry identification
  [entryset, coeff1d_est] = entryid_fct(N, s, d, basis_types, 'indices', v_i, entryid_fct_params);
  for j=1:d
    if isempty(entryset{j})
      entryset{j} = (0:N-1)';
      if show_warning
        warning('iteration %d: entryset{%d} is empty, using fallback', t, j);
      end
    end
  end
  %pairing
  [Om, coeff_est] = pairing_fct(basis_types, entryset, nodes_w_pairing, v_p, 2*s);
  %merge supports
  [T, idxOm, idxSupp]=union(Om, supp, 'rows'); % at most 3s by d
  %solving the least squares problem
  phi_t = basis_eval(basis_types, T, nodes_ls);
  phi_t = phi_t./sqrt(m_bar);%normalization
  b = phi_t\samples_ls / sqrt(m_bar); %solving ls problem, alternative methods : conjugate gradient, richardson, etc
  %prune to obtain the next approximation
  [~,sortIndexb] = sort(abs(b),'descend');
  tt=min(s,size(b,1));
  a_old = a;
  supp_old = supp;
  residual_norm_old = residual_norm;
  a=b(sortIndexb(1:tt)); % s by 1
  %new approximation's support
  supp = T(sortIndexb(1:tt),:); % s by d
  % updating current samples
  v_p=samples_pairing; %initilization
  v_i=samples_entryid; %initilization
  for jj=1:d
    v_i_jj = v_i{jj};
    nodes_entryid_jj = nodes_entryid{jj};
    v_i_jj_diff = basis_eval_mult(basis_types, supp, nodes_entryid_jj, a(:,1));
    v_i_jj = v_i_jj - v_i_jj_diff;
    v_i{jj} = v_i_jj;
  end
  
  for jj=2:d
    wz_jj = nodes_wz_pairing{jj-1};
    [m_1, m_2] = size(v_p{jj-1});
    v_p_jj_diff = basis_eval_mult(basis_types, supp, wz_jj, a(:,1));
    v_p_jj = reshape(v_p{jj-1},1,m_1*m_2);
    v_p_jj = v_p_jj - v_p_jj_diff.';
    v_p{jj-1} = reshape(v_p_jj,m_1,m_2);
  end
  
  basisEval_j_t_mul_a = basis_eval_mult(basis_types, supp, nodes_ls, a);
  residual_norm=norm(samples_ls - basisEval_j_t_mul_a); % l2 norm of residual vector
  
  if residual_norm^2 > residual_norm_old^2
    a = a_old;
    supp = supp_old;
    t = t-1;
    break;
  end
end

